@@ -567,7 +567,7 @@ get_workspace_rect (WnckPager    *pager,
       row = space % spaces_per_row;
 
       if (gtk_widget_get_direction (widget) == GTK_TEXT_DIR_RTL)
-        col = spaces_per_row - col - 1;
+        col = pager->priv->n_rows - col - 1;
 
       rect->x = (rect->width + 1) * col; 
       rect->y = (rect->height + 1) * row;
@@ -938,9 +938,9 @@ wnck_pager_draw_workspace (WnckPager    *pager,
                                      GDK_RGB_DITHER_MAX,
                                      0, 0);
     }
-  else
+  else if (active_space)
     {
-      int vx, vy, vw, vh; /* viewport */
+      double vx, vy, vw, vh; /* viewport */
       double width_ratio, height_ratio;
       cairo_t *cr;
       
@@ -1331,12 +1331,18 @@ wnck_pager_drag_data_received (GtkWidget          *widget,
 
   if ((selection_data->length != sizeof (gulong)) ||
       (selection_data->format != 8))
-    gtk_drag_finish (context, FALSE, FALSE, time);
+    {
+      gtk_drag_finish (context, FALSE, FALSE, time);
+      return;
+    }
   
   i = workspace_at_point (pager, x, y, NULL, NULL);
   space = wnck_screen_get_workspace (pager->priv->screen, i);
   if (!space)
-    gtk_drag_finish (context, FALSE, FALSE, time);
+    {
+      gtk_drag_finish (context, FALSE, FALSE, time);
+      return;
+    }
   
   xid = *((gulong *)selection_data->data);
 	      
@@ -1366,7 +1372,7 @@ wnck_pager_drag_data_get (GtkWidget        *widget,
   WnckPager *pager = WNCK_PAGER (widget);
   gulong xid;
 
-  if (pager->priv->drag_window == NULL) 
+  if (pager->priv->drag_window == NULL)
     return;
 
   xid = wnck_window_get_xid (pager->priv->drag_window);
@@ -1403,8 +1409,8 @@ wnck_pager_drag_motion_leave (GtkWidget          *widget,
 static void
 wnck_drag_clean_up (WnckWindow     *window,
 		    GdkDragContext *context,
-		    gboolean	    clean_up_window,
-		    gboolean	    clean_up_context);
+		    gboolean	    clean_up_for_context_destroy,
+		    gboolean	    clean_up_for_window_destroy);
 
 static void
 wnck_drag_context_destroyed (gpointer  windowp,
@@ -1424,9 +1430,14 @@ wnck_update_drag_icon (WnckWindow     *window,
   GtkWidget *widget;
 
   widget = g_object_get_data (G_OBJECT (context), "wnck-drag-source-widget");
+  if (!widget)
+    return;
+
   if (!gtk_icon_size_lookup_for_settings (gtk_widget_get_settings (widget),
 					  GTK_ICON_SIZE_DND, &dnd_w, &dnd_h))
     dnd_w = dnd_h = 32;
+  /* windows are huge, so let's make this huge */
+  dnd_w *= 3;
  
   workspace = wnck_window_get_workspace (window);
   if (workspace == NULL)
@@ -1437,11 +1448,13 @@ wnck_update_drag_icon (WnckWindow     *window,
   wnck_window_get_geometry (window, NULL, NULL, &org_w, &org_h);
 
   rect.x = rect.y = 0;
-  /* windows are huge, so let's make this huge */
-  dnd_w *= 3;
-  rect.width = (dnd_w * org_w) / wnck_workspace_get_width (workspace);
+  rect.width = 0.5 + ((double) (dnd_w * org_w) / (double) wnck_workspace_get_width (workspace));
   rect.width = MIN (org_w, rect.width);
-  rect.height = (rect.width * org_h) / org_w;
+  rect.height = 0.5 + ((double) (rect.width * org_h) / (double) org_w);
+
+  /* we need at least three pixels to draw the smallest window */
+  rect.width = MAX (rect.width, 3);
+  rect.height = MAX (rect.height, 3);
 
   pixmap = gdk_pixmap_new (GTK_WIDGET (widget)->window,
                            rect.width, rect.height, -1);
@@ -1462,11 +1475,13 @@ wnck_drag_window_destroyed (gpointer  contextp,
 {
   wnck_drag_clean_up ((WnckWindow *) window, GDK_DRAG_CONTEXT (contextp),
                       FALSE, TRUE);
-  /* FIXME: I'd like to have a "window is destroyed" indication, but no idea
-   * how best to do it.
-   * I didn't like setting GTK_STOCK_MISSING_IMAGE for example, because
-   * that indicates a missing image, not a missing window */
-  gtk_drag_set_icon_default (contextp);
+}
+
+static void
+wnck_drag_source_destroyed (gpointer  contextp,
+                            GObject  *drag_source)
+{
+  g_object_steal_data (G_OBJECT (contextp), "wnck-drag-source-widget");
 }
 
 /* CAREFUL: This function is a bit brittle, because the pointers given may be
@@ -1474,11 +1489,19 @@ wnck_drag_window_destroyed (gpointer  contextp,
 static void
 wnck_drag_clean_up (WnckWindow     *window,
 		    GdkDragContext *context,
-		    gboolean	    clean_up_window,
-		    gboolean	    clean_up_context)
+		    gboolean	    clean_up_for_context_destroy,
+		    gboolean	    clean_up_for_window_destroy)
 {
-  if (clean_up_window)
+  if (clean_up_for_context_destroy)
     {
+      GtkWidget *drag_source;
+
+      drag_source = g_object_get_data (G_OBJECT (context),
+                                       "wnck-drag-source-widget");
+      if (drag_source)
+        g_object_weak_unref (G_OBJECT (drag_source),
+                             wnck_drag_source_destroyed, context);
+
       g_object_weak_unref (G_OBJECT (window),
                            wnck_drag_window_destroyed, context);
       if (g_signal_handlers_disconnect_by_func (window,
@@ -1486,7 +1509,7 @@ wnck_drag_clean_up (WnckWindow     *window,
 	g_assert_not_reached ();
     }
 
-  if (clean_up_context)
+  if (clean_up_for_window_destroy)
     {
       g_object_steal_data (G_OBJECT (context), "wnck-drag-source-widget");
       g_object_weak_unref (G_OBJECT (context),
@@ -1498,7 +1521,8 @@ wnck_drag_clean_up (WnckWindow     *window,
  * wnck_window_set_as_drag_icon:
  * @window: #WnckWindow to use as drag icon
  * @context: #GdkDragContext to set the icon on
- * @drag_source: #GtkWidget that started the drag
+ * @drag_source: #GtkWidget that started the drag, or one of its parent. This
+ * widget needs to stay alive as long as possible during the drag.
  *
  * Sets the given @window as the drag icon for @context.
  **/
@@ -1517,6 +1541,8 @@ _wnck_window_set_as_drag_icon (WnckWindow     *window,
                     G_CALLBACK (wnck_update_drag_icon), context);
 
   g_object_set_data (G_OBJECT (context), "wnck-drag-source-widget", drag_source);
+  g_object_weak_ref (G_OBJECT (drag_source), wnck_drag_source_destroyed, context);
+
   g_object_weak_ref (G_OBJECT (context), wnck_drag_context_destroyed, window);
 
   wnck_update_drag_icon (window, context);
