@@ -68,10 +68,13 @@ struct _WnckScreenPrivate
   GList *workspaces;
 
   /* previously_active_window is used in tandem with active_window to
-   * determine return status of meta_window_is_most_recently_actived().
+   * determine return status of wnck_window_is_most_recently_actived().
+   * These are usually shared for all screens, although this is not guaranteed
+   * to be true.
    */
   WnckWindow *active_window;
   WnckWindow *previously_active_window;
+
   WnckWorkspace *active_workspace;
 
   /* Provides the sorting order number for the next window, to make
@@ -178,7 +181,7 @@ wnck_screen_get_type (void)
   
   if (!object_type)
     {
-      static const GTypeInfo object_info =
+      const GTypeInfo object_info =
       {
         sizeof (WnckScreenClass),
         (GBaseInitFunc) NULL,
@@ -368,6 +371,7 @@ static void
 wnck_screen_finalize (GObject *object)
 {
   WnckScreen *screen;
+  gpointer weak_pointer;
 
   screen = WNCK_SCREEN (object);
   
@@ -381,6 +385,18 @@ wnck_screen_finalize (GObject *object)
   g_list_free (screen->priv->stacked_windows);
   g_list_free (screen->priv->workspaces);
 
+  weak_pointer = &screen->priv->active_window;
+  if (screen->priv->active_window != NULL)
+    g_object_remove_weak_pointer (G_OBJECT (screen->priv->active_window),
+                                  weak_pointer);
+  screen->priv->active_window = NULL;
+
+  weak_pointer = &screen->priv->previously_active_window;
+  if (screen->priv->previously_active_window != NULL)
+    g_object_remove_weak_pointer (G_OBJECT (screen->priv->previously_active_window),
+                                  weak_pointer);
+  screen->priv->previously_active_window = NULL;
+
   screens[screen->priv->number] = NULL;
 
 #ifdef HAVE_STARTUP_NOTIFICATION
@@ -1077,6 +1093,48 @@ wnck_screen_free_workspace_layout (WnckWorkspaceLayout *layout)
   g_free (layout->grid);
 }
 
+static void
+set_active_window (WnckScreen *screen,
+                   WnckWindow *window)
+{
+  gpointer weak_pointer;
+
+  weak_pointer = &screen->priv->active_window;
+
+  /* we need the weak pointer since the active window might be shared between
+   * two screens, and so the value for one screen might become invalid when
+   * the window is destroyed on another screen */
+  if (screen->priv->active_window != NULL)
+    g_object_remove_weak_pointer (G_OBJECT (screen->priv->active_window),
+                                  weak_pointer);
+
+  screen->priv->active_window = window;
+  if (screen->priv->active_window != NULL)
+    g_object_add_weak_pointer (G_OBJECT (screen->priv->active_window),
+                               weak_pointer);
+}
+
+static void
+set_previously_active_window (WnckScreen *screen,
+                              WnckWindow *window)
+{
+  gpointer weak_pointer;
+
+  weak_pointer = &screen->priv->previously_active_window;
+
+  /* we need the weak pointer since the active window might be shared between
+   * two screens, and so the value for one screen might become invalid when
+   * the window is destroyed on another screen */
+  if (screen->priv->previously_active_window != NULL)
+    g_object_remove_weak_pointer (G_OBJECT (screen->priv->previously_active_window),
+                                  weak_pointer);
+
+  screen->priv->previously_active_window = window;
+  if (screen->priv->previously_active_window != NULL)
+    g_object_add_weak_pointer (G_OBJECT (screen->priv->previously_active_window),
+                               weak_pointer);
+}
+
 static gboolean
 lists_equal (GList *a,
              GList *b)
@@ -1382,18 +1440,18 @@ update_client_list (WnckScreen *screen)
 
       window = WNCK_WINDOW (tmp->data);
 
+      if (window == screen->priv->previously_active_window)
+        {
+          set_previously_active_window (screen, NULL);
+        }
+      
       if (window == screen->priv->active_window)
         {
-          screen->priv->previously_active_window = screen->priv->active_window;
-          screen->priv->active_window = NULL;
+          set_previously_active_window (screen, screen->priv->active_window);
+          set_active_window (screen, NULL);
           active_changed = TRUE;
         }
 
-      if (window == screen->priv->previously_active_window)
-        {
-          screen->priv->previously_active_window = NULL;
-        }
-      
       emit_window_closed (screen, window);
     }
 
@@ -1425,6 +1483,8 @@ update_client_list (WnckScreen *screen)
   g_list_free (created);
   g_list_free (closed_apps);
   g_list_free (created_apps);
+  g_list_free (closed_class_groups);
+  g_list_free (created_class_groups);
 
   --reentrancy_guard;
 
@@ -1715,8 +1775,8 @@ update_active_window (WnckScreen *screen)
   if (window == screen->priv->active_window)
     return;
 
-  screen->priv->previously_active_window = screen->priv->active_window;
-  screen->priv->active_window = window;
+  set_previously_active_window (screen, screen->priv->active_window);
+  set_active_window (screen, window);
 
   emit_active_window_changed (screen);
 }
